#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Indresh Vishwakarma.
# Distributed under the terms of the Modified BSD License.

from .maze import Maze
from ._version import __version__, version_info

def _jupyter_labextension_paths():
    """Called by Jupyter Lab Server to detect if it is a valid labextension and
    to install the widget
    Returns
    =======
    src: Source directory name to copy files from. Webpack outputs generated files
        into this directory and Jupyter Lab copies from this directory during
        widget installation
    dest: Destination directory name to install widget files to. Jupyter Lab copies
        from `src` directory into <jupyter path>/labextensions/<dest> directory
        during widget installation
    """
    return [{
        'src': 'labextension',
        'dest': 'ttgtcanvas2',
    }]


def _jupyter_nbextension_paths():
    """Called by Jupyter Notebook Server to detect if it is a valid nbextension and
    to install the widget
    Returns
    =======
    section: The section of the Jupyter Notebook Server to change.
        Must be 'notebook' for widget extensions
    src: Source directory name to copy files from. Webpack outputs generated files
        into this directory and Jupyter Notebook copies from this directory during
        widget installation
    dest: Destination directory name to install widget files to. Jupyter Notebook copies
        from `src` directory into <jupyter path>/nbextensions/<dest> directory
        during widget installation
    require: Path to importable AMD Javascript module inside the
        <jupyter path>/nbextensions/<dest> directory
    """
    return [{
        'section': 'notebook',
        'src': 'nbextension',
        'dest': 'ttgtcanvas2',
        'require': 'ttgtcanvas2/extension'
    }]



_directions = [(1,0), (0,1), (-1,0), (0,-1)]

dir_names = ["east", "north","west", "south"]

class Robot():
    def __init__(self, index, robot_model, world):
        self.model = robot_model
        self.world_model = world.model
        self.world = world
        self.index = index
        self.collections = {}
 
    @property
    def x(self):
        return self.model.x
    
    @property
    def y(self):
        return self.model.y
    
    @property
    def dir(self):
        return self.model.orientation
    
    @dir.setter
    def dir(self, value):
        self.model.orientation = value

    @x.setter
    def x(self,value):
        self.model.x = value
    
    @y.setter
    def y(self, value):
        self.model.y = value
    
    def move(self, step=1):
        for x in range(step):
            self._move()
        self.world.js_call('move_to',[self.index, self.x, self.y])
    
    def _move(self):
        if self.front_is_clear():
            xx, yy = _directions[self.dir]
            self.x += xx
            self.y += yy
        else: 
            self.world.js_call('error', ['Opps You Hit the wall'])
    
    def front_is_clear(self):
        return self.world_model.is_clear(self.x, self.y, self.dir)
    
    def turn_left(self):
        self.dir = (self.dir + 1) % 4
        self.world.js_call('turn_left', [self.index])
    
    def is_front_clear(self):
        return self.world_model.is_clear(self.x, self.y, self.dir)
    
    def wall_in_front(self):
        return not self.is_front_clear()

    def build_wall(self):
        if self.world_model.is_clear(self.x, self.y, self.dir):
            if self.dir == 0:
                self.world_model.add_wall(self.x, self.y, "east")
            if self.dir == 1:
                self.world_model.add_wall(self.x, self.y, "north")
            if self.dir == 2:
                self.world_model.add_wall(self.x - 1, self.y, "east")
            if self.dir == 3:
                self.world_model.add_wall(self.x, self.y - 1, "north")
        else:
            self.world.js_call("error", ["Wall already exists"])

    def set_trace(self, color='red'):
        self.world.js_call('set_trace', [self.index, color])
    
    def on_object(self):
        k = "{},{}".format(self.x,self.y)
        obj  = self.world_model.objects.get(k, None)
        return  obj is not None and obj != self.collections.get(k, None)
    
    def pick(self):
        if self.on_object():
            ck = "{},{}".format(self.x,self.y)
        
            obj = self.world_model.objects.get(ck, None)
            if obj is not None:
                for k, val in obj.items():
                    collection = self.collections.get(ck, {})
                    picked = collection.get(k, 0)
                    picked = picked + 1
                    collection[k] = picked
                    self.collections[ck] = collection
                    if val == picked:
                        self.world.js_call('remove_object', [self.x,self.y])
                    if val > picked:
                        self.world.js_call('update_object', [self.x, self.y, val - 1])
        else:
            self.world.js_call('error', ["No Items to Pick"])




from .robot_model import RobotModel

class WorldModel():
    def __init__(self, ui,kwargs):
        self.walls = kwargs.get('walls', {})
        self.robots_config = kwargs.get('robots', [])
        self.goals = kwargs.get('goal', {})
        self.objects = kwargs.get('objects', {})
        self.ui = ui
        self.added_walls = {}
        self.errors = []
        self.set_dimensions(kwargs)
        self.init_robots(self.robots_config)


    def init_robots(self, robots=[]):
        self.robots = [RobotModel(robot) for robot in robots]

    def set_dimensions(self, options={}):
        self.rows = options.get('rows',10)
        self.cols = options.get('cols', 10)
    
    def dimensions(self):
        return [self.rows, self.cols]
    
    def is_clear(self, x, y, dir):
        return not self.is_wall(x,y, dir) and not self.is_border(x, y, dir)
    
    def add_wall(self, x, y, dir):
        key = "{},{}".format(x,y)
        val = self.walls.get(key,[])
        val.append(dir)
        self.walls[key] = val
        self.added_walls[key] = val
        self.ui.js_call('add_wall', [x,y, dir])

    
    def is_wall(self, x, y, dir):
        wall =  self.walls.get("{},{}".format(x,y), [])
        if dir == 0 and "east" in wall:
            return True 
        if dir == 1 and "north" in wall: #north
            return True 

        wall =  self.walls.get("{},{}".format(x - 1 , y), [])
        if dir == 2 and "east" in wall:
            return True 

        wall =  self.walls.get("{},{}".format(x,y -1), [])
        if dir == 3 and "north" in wall: #south
            return True 
        return False

    def is_border(self, x,y, dir):
        if dir == 0 and x == self.cols : #east
            return True 
        if dir == 1 and y == self.rows : #north
            return True 
        if dir == 2 and x - 1 == 0 : #west
            return True 
        if dir == 3 and y - 1 == 0 : #south
            return True 
        return False
    
    def check(self, bot): 
      return self.valid_position(bot) and self.all_walls_builded(bot) and self.all_objects_collected(bot)

    def valid_position(self, bot):
        final_position = self.goals.get('possible_final_positions', self.goals.get('position', None))
        if final_position is None: 
            return True
        
        if isinstance(final_position, dict):
            return final_position.get("x") == bot.x and final_position.get("y") == bot.y
        else:
            coord = [bot.x, bot.y]
            return coord in final_position
    

    def all_walls_builded(self, bot):
        walls = self.goals.get('walls', None)
        if walls is None: 
            return True
        return all(sorted(self.added_walls.get(k, [])) == sorted(v) for k, v in walls.items())
    
    def all_objects_collected(self, bot):
        objects = self.goals.get('objects', None)
        if objects is None:
            return True
        return bot.collections == objects
            


    




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Indresh Vishwakarma.
# Distributed under the terms of the Modified BSD License.

from ipywidgets import DOMWidget
from traitlets import Unicode
from ._frontend import module_name, module_version

import time as _time
from datetime import datetime
import json



from .models import WorldModel, RobotModel
from .robot import Robot


class Maze(DOMWidget):
    """TODO: Add docstring here
    """
    _model_name = Unicode('MazeModel').tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _view_name = Unicode('MazeView').tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)

    current_call  = Unicode('{}').tag(sync=True)
    method_return = Unicode('{}').tag(sync=True)


    def js_call(self, method_name, params): 
        # print("calling method: " + method_name)
        cb = datetime.now().strftime('%f')
        self.current_call = json.dumps({'method_name': method_name, 'params': params, 'cb': cb})
    
    def __init__(self, kwargs):
        super(Maze, self).__init__()
        self.options = kwargs
        self.model = WorldModel(self, kwargs)
        self.robots = [Robot(idx, x, self) for idx, x in enumerate(self.model.robots)]


        display(self)
        _time.sleep(1)

        self.js_call('draw_all', [
            self.model.rows, 
            self.model.cols, 
            self.model.walls,
            self.options.get("robots", []),
            self.model.objects
        ])
    
    def bot(self, bot_index=0):
        return self.robots[bot_index]
    
    def check(self, bot_index=0):
        if self.model.check(self.bot(bot_index)):
            self.js_call('msg', ['Task Completed'])
        else:
            self.js_call('error', ["One Or More Goal are Not Completed."])




import random
class RobotModel():
    def __init__(self, kwargs):
        self.x = kwargs.get("x", 1)
        self.y = kwargs.get("y", 1)
        self.orientation = kwargs.get('_orientation', 0)
        self.traceColor = kwargs.get('_trace_color', None)
        self.possible_initial_positions = kwargs.get('possible_initial_positions', None)
        if self.possible_initial_positions is not None: 
            pos = random.choice(self.possible_initial_positions)
            self.x = pos[0]
            self.y = pos[1]